from numpy.testing import assert_almost_equal, assert_equal
from pylj import forcefields
import unittest


class TestForcefields(unittest.TestCase):
    def test_lennard_jones_energy(self):
        a = forcefields.lennard_jones(2.0, [1.0, 1.0])
        assert_almost_equal(a, -0.015380859)
        b = forcefields.lennard_jones([2.0, 4.5], [1.0, 1.0])
        assert_almost_equal(b, [-0.015380859, -0.00012041])
        c = forcefields.lennard_jones([2.0, 4.5], [0.5, 3.5])
        assert_almost_equal(c, [-0.05456543, -0.00042149])
        d = forcefields.lennard_jones([1.0, 1.5, 20.0], [5.0, 3.5])
        assert_almost_equal(
            d, [1.50000000, -0.268733500, -5.46874988e-08])
        e = forcefields.lennard_jones([100.0, 200.0, 500.0], [100.0, 300.0])
        assert_almost_equal(e, [0, 0, 0])

    def test_lennard_jones_force(self):
        a = forcefields.lennard_jones(2.0, [1.0, 1.0], force=True)
        assert_almost_equal(a, -0.045410156)
        b = forcefields.lennard_jones([2.0, 4.0, 6.0], [1.0, 1.0], force=True)
        assert_almost_equal(b, [-0.045410156, -3.66032124e-04, -2.14325517e-05])
        c = forcefields.lennard_jones([2.0, 4.0, 6.0], [1.5, 4.0], force=True)
        assert_almost_equal(c, [-0.185302734, -1.46457553e-03, -8.57325038e-05])
        d = forcefields.lennard_jones(
            [150.0, 300.0, 500.0], [200.0, 500.0], force=True)
        assert_almost_equal(d, [-1.7558299e-12, -1.3717421e-14, -3.8400000e-16])

    def test_lennard_jones_sigma_epsilon_energy(self):
        a = forcefields.lennard_jones_sigma_epsilon(2.0, [1.0, 0.25])
        assert_almost_equal(a, -0.015380859)
        b = forcefields.lennard_jones_sigma_epsilon([2.0, 1.0], [1.0, 0.25])
        assert_almost_equal(b, [-0.015380859, 0])
        c = forcefields.lennard_jones_sigma_epsilon(
            [2.0, 1.0, 1.5], [0.5, 0.75])
        assert_almost_equal(c, [0, 2.953125, 0.0190068])
        d = forcefields.lennard_jones_sigma_epsilon(
            [400.0, 500.0, 600.0], [5e-10, 9e-9])
        assert_almost_equal(d, [0, 0, 0])

    def test_lennard_jones_sigma_epsilon_force(self):
        a = forcefields.lennard_jones_sigma_epsilon(
            2.0, [1.0, 0.25], force=True)
        assert_almost_equal(a, -0.045410156)
        b = forcefields.lennard_jones_sigma_epsilon(
            [2.0, 4.0], [1.0, 0.25], force=True)
        assert_almost_equal(
            b, [-0.0454102, -0.000366])
        c = forcefields.lennard_jones_sigma_epsilon(
            [3.0, 4.0], [3.0, 1.0], force=True)
        assert_almost_equal(c, [8.0, -0.6877549])

    def test_buckingham_energy(self):
        a = forcefields.buckingham(2.0, [1.0, 1.0, 1.0])
        assert_almost_equal(a, 0.1197103832)
        b = forcefields.buckingham([2.0], [1.0, 1.0, 1.0])
        assert_almost_equal(b, 0.1197103832)
        c = forcefields.buckingham([2.0, 4.0], [1.0, 1.0, 1.0])
        assert_almost_equal(c, [0.1197103832, 0.0180715])
        d = forcefields.buckingham([2.0, 4.0, 5.0], [0.01, 0.01, 0.01])
        assert_almost_equal(d, [0.0096457, 0.0096055, 0.0095117])

    def test_buckingham_force(self):
        a = forcefields.buckingham(2.0, [1.0, 1.0, 1.0], force=True)
        assert_almost_equal(a, 0.08846028324)
        b = forcefields.buckingham([2.0], [1.0, 1.0, 1.0], force=True)
        assert_almost_equal(b, 0.08846028324)
        c = forcefields.buckingham(
            [2.0, 1.0, 4.0], [1.5, 0.1, 2.0], force=True)
        assert_almost_equal(c, [0.0290596, -11.8642744, 0.0998156])

    def test_square_well_energy(self):
        a = forcefields.square_well(2.0, [1.0, 1.5, 2.0])
        assert_equal(a, -1.0)
        b = forcefields.square_well(0.5, [1.0, 2.0, 1.25])
        assert_equal(b, float('inf'))
        c = forcefields.square_well(3.0, [0.5, 1.5, 1.25])
        assert_equal(c, 0)
        d = forcefields.square_well([2.0, 0.5], [1.0, 1.5, 2.0])
        assert_equal(d, [-1.0, float('inf')])
        e = forcefields.square_well([3.0, 3.0, 0.25], [1.0, 1.5, 1.25])
        assert_equal(e, [0, 0, float('inf')])
        f = forcefields.square_well(
            [3.0, 3.0, 0.25], [1.0, 1.5, 1.25], max_val=5000)
        assert_equal(f, [0, 0, 5000])

    def test_square_well_force(self):
        with self.assertRaises(ValueError):
            forcefields.square_well(
                2.0, [1.0, 1.5, 2.0], force=True)
        with self.assertRaises(ValueError):
            forcefields.square_well(
                [2.0], [1.0, 1.5, 2.0], force=True)


if __name__ == '__main__':
    unittest.main(exit=False)



import numpy as np
from numba import jit


@jit
def lennard_jones(dr, constants, force=False):
    r"""Calculate the energy or force for a pair of particles using the
    Lennard-Jones (A/B variant) forcefield.

    .. math::
        E = \frac{A}{dr^{12}} - \frac{B}{dr^6}

    .. math::
        f = \frac{12A}{dr^{13}} - \frac{6B}{dr^7}

    Parameters
    ----------
    dr: float, array_like
        The distances between the all pairs of particles.
    constants: float, array_like
        An array of length two consisting of the A and B parameters for the
        12-6 Lennard-Jones function
    force: bool (optional)
        If true, the negative first derivative will be found.

    Returns
    -------
    float: array_like
        The potential energy or force between the particles.
    """

    if force:
        return 12 * constants[0] * np.power(dr, -13) - (
            6 * constants[1] * np.power(dr, -7))

    else:
        return constants[0] * np.power(dr, -12) - (
            constants[1] * np.power(dr, -6))


@jit
def lennard_jones_sigma_epsilon(dr, constants, force=False):
    r"""Calculate the energy or force for a pair of particles using the
    Lennard-Jones (sigma/epsilon variant) forcefield.

    .. math::
        E = \frac{4e*a^{12}}{dr^{12}} - \frac{4e*a^{6}}{dr^6}

    .. math::
        f = \frac{48e*a^{12}}{dr^{13}} - \frac{24e*a^{6}}{dr^7}

    Parameters
    ----------
    dr: float, array_like
        The distances between the all pairs of particles.
    constants: float, array_like
        An array of length two consisting of the sigma (a) and epsilon (e)
        parameters for the 12-6 Lennard-Jones function
    force: bool (optional)
        If true, the negative first derivative will be found.

    Returns
    -------
    float: array_like
        The potential energy or force between the particles.
    """

    if force:
        return 48 * constants[1] * np.power(constants[0], 12) * np.power(
            dr, -13) - (24 * constants[1] * np.power(
                constants[0], 6) * np.power(dr, -7))
    else:
        return 4 * constants[1] * np.power(dr, -12) - (
            4 * constants[1] * np.power(constants[0], 6) * np.power(dr, -6))


@jit
def buckingham(dr, constants, force=False):
    r""" Calculate the energy or force for a pair of particles using the
    Buckingham forcefield.

    .. math::
        E = Ae^{(-Bdr)} - \frac{C}{dr^6}

    .. math::
        f = ABe^{(-Bdr)} - \frac{6C}{dr^7}

    Parameters
    ----------
    dr: float, array_like
        The distances between all the pairs of particles.
    constants: float, array_like
        An array of length three consisting of the A, B and C parameters for
        the Buckingham function.
    force: bool (optional)
        If true, the negative first derivative will be found.

    Returns
    -------
    float: array_like
        The potential energy or force between the particles.
    """
    if force:
        return constants[0] * constants[1] * np.exp(
            - np.multiply(constants[1], dr)) - 6 * constants[2] / np.power(
            dr, 7)
    else:
        return constants[0] * np.exp(
            - np.multiply(constants[1], dr)) - constants[2] / np.power(dr, 6)


def square_well(dr, constants, max_val=np.inf, force=False):
    r'''Calculate the energy or force for a pair of particles using a
    square well model.

    .. math::
        E = {
        if dr < sigma:
            E = max_val
        elif sigma <= dr < lambda * sigma:
            E = -epsilon
        elif r >= lambda * sigma:
            E = 0
        }
    .. math::
        f = {
        if sigma <= dr < lambda * sigma:
            f = inf
        else:
            f = 0
        }
    Parameters
    ----------
    dr: float, array_like
        The distances between all the pairs of particles.
    constants: float, array_like
        An array of length three consisting of the epsilon, sigma, and lambda
        parameters for the square well model.
    max_val: int (optional)
        Upper bound for values in square well - replaces usual infinite values
    force: bool (optional)
        If true, the negative first derivative will be found.

    Returns
    -------
    float: array_like
        The potential energy between the particles.
    '''
    if not isinstance(dr, np.ndarray):
        if isinstance(dr, list):
            dr = np.array(dr, dtype='float')
        elif isinstance(dr, float):
            dr = np.array([dr], dtype='float')

    if force:
        raise ValueError("Force is infinite at sigma <= dr < lambda * sigma")

    else:
        E = np.zeros_like(dr)
        E[np.where(dr < constants[0])] = max_val
        E[np.where(dr >= constants[2] * constants[1])] = 0

        # apply mask for sigma <= dr < lambda * sigma
        a = constants[1] <= dr
        b = dr < constants[2] * constants[1]
        E[np.where(a & b)] = -constants[0]

        if len(E) == 1:
            return float(E[0])
        else:
            return np.array(E, dtype='float')



from numpy.testing import assert_almost_equal, assert_equal
from pylj import md
import unittest


class TestMd(unittest.TestCase):
    def test_initialise_square(self):
        a = md.initialise(2, 300, 8, "square")
        assert_equal(a.number_of_particles, 2)
        assert_almost_equal(a.box_length, 8e-10)
        assert_almost_equal(a.init_temp, 300)
        assert_almost_equal(a.particles["xposition"] * 1e10, [2, 2])
        assert_almost_equal(a.particles["yposition"] * 1e10, [2, 6])

    def test_velocity_verlet(self):
        a = md.initialise(2, 300, 8, "square")
        a.particles, a.distances, a.forces, a.energies = md.velocity_verlet(
            a.particles, 1, a.box_length, a.cut_off, a.constants, a.forcefield, a.mass
        )
        assert_almost_equal(a.particles["xprevious_position"] * 1e10, [2, 2])
        assert_almost_equal(a.particles["yprevious_position"] * 1e10, [2, 6])

    def test_update_positions(self):
        a = md.initialise(2, 300, 8, "square")
        a.particles["xvelocity"] = 1e4
        a.particles["yvelocity"] = 1e4
        a.particles["xacceleration"] = 1e4
        a.particles["yacceleration"] = 1e4
        b, c = md.update_positions(
            [a.particles["xposition"], a.particles["yposition"]],
            [a.particles["xprevious_position"], a.particles["yprevious_position"]],
            [a.particles["xvelocity"], a.particles["yvelocity"]],
            [a.particles["xacceleration"], a.particles["yacceleration"]],
            a.timestep_length,
            a.box_length,
        )
        assert_almost_equal(b[0][0] * 1e10, 3)
        assert_almost_equal(b[1][0] * 1e10, 3)
        assert_almost_equal(b[0][1] * 1e10, 3)
        assert_almost_equal(b[1][1] * 1e10, 7)

    def test_update_velocities(self):
        a = md.initialise(2, 300, 8, "square")
        a.particles["xvelocity"] = 1e-10
        a.particles["yvelocity"] = 1e-10
        a.particles["xacceleration"] = 1e4
        a.particles["yacceleration"] = 1e4
        xacceleration_new = 2e4
        yacceleration_new = 2e4
        b = md.update_velocities(
            [a.particles["xvelocity"], a.particles["yvelocity"]],
            [xacceleration_new, yacceleration_new],
            [a.particles["xacceleration"], a.particles["yacceleration"]],
            a.timestep_length,
        )
        assert_almost_equal(b[0][0] * 1e10, 2.5)
        assert_almost_equal(b[1][0] * 1e10, 2.5)
        assert_almost_equal(b[0][1] * 1e10, 2.5)
        assert_almost_equal(b[1][1] * 1e10, 2.5)

    def test_calculate_temperature(self):
        a = md.initialise(1, 300, 8, "square")
        a.particles["xvelocity"] = [1e-10]
        a.particles["yvelocity"] = [1e-10]
        a.particles["xacceleration"] = [1e4]
        a.particles["yacceleration"] = [1e4]
        b = md.calculate_temperature(a.particles, mass=39.948)
        assert_almost_equal(b * 1e23, 4.8048103702737945)

    def test_calculate_msd(self):
        a = md.initialise(2, 300, 8, "square")
        a.particles["xposition"] = [3e-10, 3e-10]
        a.particles["yposition"] = [3e-10, 7e-10]
        b = md.calculate_msd(a.particles, a.initial_particles, a.box_length)
        assert_almost_equal(b, 2e-20)

    def test_calculate_msd_large(self):
        a = md.initialise(2, 300, 8, "square")
        a.particles["xposition"] = [7e-10, 3e-10]
        a.particles["yposition"] = [7e-10, 7e-10]
        b = md.calculate_msd(a.particles, a.initial_particles, a.box_length)
        assert_almost_equal(b, 10e-20)



import numpy as np
from pylj import forcefields as ff


def initialise(
    number_of_particles,
    temperature,
    box_length,
    init_conf,
    mass=39.948,
    constants=[1.363e-134, 9.273e-78],
    forcefield=ff.lennard_jones,
):
    """Initialise the particle positions (this can be either as a square or
    random arrangement), velocities (based on the temperature defined, and #
    calculate the initial forces/accelerations.

    Parameters
    ----------
    number_of_particles: int
        Number of particles to simulate.
    temperature: float
        Initial temperature of the particles, in Kelvin.
    box_length: float
        Length of a single dimension of the simulation square, in Angstrom.
    init_conf: string, optional
        The way that the particles are initially positioned. Should be one of:
        - 'square'
        - 'random'
    mass: float (optional)
        The mass of the particles being simulated.
    constants: float, array_like (optional)
        The values of the constants for the forcefield used.
    forcefield: function (optional)
        The particular forcefield to be used to find the energy and forces.

    Returns
    -------
    System
        System information.
    """
    from pylj import util

    system = util.System(
        number_of_particles,
        temperature,
        box_length,
        constants,
        forcefield,
        mass,
        init_conf=init_conf,
    )
    system.particles["xvelocity"] = 0
    system.particles["yvelocity"] = 0
    return system


def initialize(number_particles, temperature, box_length, init_conf):
    """Maps to the mc.initialise function to account for US english spelling.
    """
    a = initialise(number_particles, temperature, box_length, init_conf)
    return a


def sample(total_energy, system):
    """Sample parameters of interest in the simulation.

    Parameters
    ----------
    total_energy: float
        The total system energy.
    system: System
        Details about the whole system

    Returns
    -------
    System:
        Details about the whole system, with the new temperature, pressure,
        msd, and force appended to the appropriate
        arrays.
    """
    system.energy_sample = np.append(system.energy_sample, total_energy)
    return system


def select_random_particle(particles):
    """Selects a random particle from the system and return its index and
    current position.

    Parameters
    ----------
    particles: util.particle.dt, array_like
        Information about the particles.

    Returns
    -------
    int:
        Index of the random particle that is selected.
    float, array_like:
        The current position of the chosen particle.
    """
    random_particle = np.random.randint(0, particles.size)
    position_store = [
        particles["xposition"][random_particle],
        particles["yposition"][random_particle],
    ]
    return random_particle, position_store


def get_new_particle(particles, random_particle, box_length):
    """Generates a new position for the particle.

    Parameters
    ----------
    particles: util.particle.dt, array_like
        Information about the particles.
    random_particle: int
        Index of the random particle that is selected.
    box_length: float
        Length of a single dimension of the simulation square.

    Returns
    -------
    util.particle.dt, array_like
        Information about the particles, updated to account for the change of
        selected particle position.
    """
    particles["xposition"][random_particle] = np.random.uniform(0, box_length)
    particles["yposition"][random_particle] = np.random.uniform(0, box_length)
    return particles


def accept(new_energy):
    """Accept the move.

    Parameters
    ----------
    new_energy: float
        A new total energy for the system.

    Returns
    -------
    float:
        A new total energy for the system.
    """
    return new_energy


def reject(position_store, particles, random_particle):
    """Reject the move and return the particle to the original place.

    Parameters
    ----------
    position_store: float, array_like
        The x and y positions previously held by the particle that has moved.
    particles: util.particle.dt, array_like
        Information about the particles.
    random_particle: int
        Index of the random particle that is selected.

    Returns
    -------
    util.particle.dt, array_like
        Information about the particles, with the particle returned to the
        original position
    """
    particles["xposition"][random_particle] = position_store[0]
    particles["yposition"][random_particle] = position_store[1]
    return particles


def metropolis(temperature, old_energy, new_energy, n=np.random.rand()):
    """Determines if the move is accepted or rejected based on the metropolis
    condition.

    Parameters
    ----------
    temperature: float
        Simulation temperature.
    old_energy: float
        The total energy of the simulation in the previous configuration.
    new_energy: float
        The total energy of the simulation in the current configuration.
    n: float, optional
        The random number against which the Metropolis condition is tested. The
        default is from a numpy uniform distribution.

    Returns
    -------
    bool
        True if the move should be accepted.
    """
    boltzmann_constant = 1.3806e-23  # joules/kelvin
    beta = 1 / (boltzmann_constant * temperature)
    energy_difference = new_energy - old_energy
    metropolis_factor = np.exp(-beta * energy_difference)
    if n < metropolis_factor:
        return True
    else:
        return False



from numpy.testing import assert_almost_equal, assert_equal
from pylj import util
from pylj import forcefields as ff
import unittest


class TestUtil(unittest.TestCase):
    def test_system_square(self):
        a = util.System(
            2,
            300,
            8,
            mass=39.948,
            constants=[1.363e-134, 9.273e-78],
            forcefield=ff.lennard_jones,
        )
        assert_equal(a.number_of_particles, 2)
        assert_equal(a.init_temp, 300)
        assert_almost_equal(a.box_length * 1e10, 8)
        assert_almost_equal(a.timestep_length, 1e-14)
        assert_almost_equal(a.particles["xposition"] * 1e10, [2, 2])
        assert_almost_equal(a.particles["yposition"] * 1e10, [2, 6])
        assert_almost_equal(a.initial_particles["xposition"] * 1e10, [2, 2])
        assert_almost_equal(a.initial_particles["yposition"] * 1e10, [2, 6])
        assert_almost_equal(a.cut_off * 1e10, 4.0)
        assert_equal(a.distances.size, 1)
        assert_equal(a.forces.size, 1)
        assert_equal(a.energies.size, 1)

    def test_system_random(self):
        a = util.System(
            2,
            300,
            8,
            init_conf="random",
            mass=39.948,
            constants=[1.363e-134, 9.273e-78],
            forcefield=ff.lennard_jones,
        )
        assert_equal(a.number_of_particles, 2)
        assert_equal(a.init_temp, 300)
        assert_almost_equal(a.box_length * 1e10, 8)
        assert_almost_equal(a.timestep_length, 1e-14)
        self.assertTrue(0 <= a.particles["xposition"][0] * 1e10 <= 8)
        self.assertTrue(0 <= a.particles["yposition"][0] * 1e10 <= 8)
        self.assertTrue(0 <= a.particles["xposition"][1] * 1e10 <= 8)
        self.assertTrue(0 <= a.particles["yposition"][1] * 1e10 <= 8)
        assert_almost_equal(a.cut_off * 1e10, 4.0)
        assert_equal(a.distances.size, 1)
        assert_equal(a.forces.size, 1)
        assert_equal(a.energies.size, 1)

    def test_system_too_big(self):
        with self.assertRaises(AttributeError) as context:
            util.System(
                2,
                300,
                1000,
                mass=39.948,
                constants=[1.363e-134, 9.273e-78],
                forcefield=ff.lennard_jones,
            )
        self.assertTrue(
            "With a box length of 1000 the particles are probably "
            "too small to be seen in the viewer. Try something "
            "(much) less than 600." in str(context.exception)
        )

    def test_system_too_small(self):
        with self.assertRaises(AttributeError) as context:
            util.System(
                2,
                300,
                2,
                mass=39.948,
                constants=[1.363e-134, 9.273e-78],
                forcefield=ff.lennard_jones,
            )
        self.assertTrue(
            "With a box length of 2 the cell is too small to "
            "really hold more than one particle." in str(context.exception)
        )

    def test_system_init_conf(self):
        with self.assertRaises(NotImplementedError) as context:
            util.System(
                2,
                300,
                100,
                init_conf="horseradish",
                mass=39.948,
                constants=[1.363e-134, 9.273e-78],
                forcefield=ff.lennard_jones,
            )
        self.assertTrue(
            "The initial configuration type horseradish is not "
            "recognised. Available options are: square or "
            "random" in str(context.exception)
        )
